-- Stan Bak
-- Computational Element for network simulator - 12-6-08

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use work.bak.all;

entity comp_ele is
	generic	
	(	
		SELF_ADDR : std_logic_vector(7 downto 0) := (others => '0');  	
		CONNECTIONS : integer := 4
	);	   
	port 
	( 
		clock			: in std_logic;
		gen_task 		: in std_logic;
		gen_dest 		: in std_logic_vector(7 downto 0);
		read_done 		: in std_logic;
		read_ok 		: in std_logic; -- signal
		read_data 		: in std_logic_vector(47 downto 0);
		tick_time 		: in std_logic; -- should we increment the virtual time?
		logging_ack 	: in std_logic; -- logging acknowledgement
		mutex_grant 		: in std_logic_vector(CONNECTIONS-1 downto 0);
		delay			: in unsigned(8*CONNECTIONS - 1  downto 0);
		route_index 	: in integer range 0 to CONNECTIONS-1; -- connected to routing table (output)
		
		readmutex_grant	: in std_logic;
		readmutex_req		: out std_logic := '0';

		gen_ack			: out std_logic := '0';
		mutex_req 		: out std_logic_vector(CONNECTIONS-1 downto 0) := (others => '0');
		mutex_write 		: out std_logic_vector(CONNECTIONS-1 downto 0) := (others => '0'); -- (signal)
		mutex_data 		: out std_logic_vector(48*CONNECTIONS - 1 downto 0) := (others => '0'); -- ew
		read_en 		: out std_logic := '0'; -- (active high)
		logging_en 		: out std_logic := '0'; -- signal
		logging_data 	: out std_logic_vector(47 downto 0) := (others => '0');
		comp_done 		: out std_logic := '0'; -- computation done for this virutal clock tick?
		dest 			: out std_logic_vector(7 downto 0) := (others => '0') -- connected to routing table (input)
	); 
end comp_ele;

architecture Behavioral of comp_ele is

-- internal
signal dest_internal : std_logic_vector(7 downto 0) := (others => '0');
signal src : std_logic_vector(7 downto 0) := (others => '0');
signal packet_time : std_logic_vector(15 downto 0) := (others => '0');

signal curTime : std_logic_vector(15 downto 0) := (others => '0');
signal sent_time : std_logic_vector(15 downto 0) := (others => '0');

constant STATE_INIT : integer := 0;
constant STATE_DID_LOOKUP_ROUTE : integer := 1;
constant STATE_WAITMUTEX : integer := 2;
constant STATE_WROTE_PACKET : integer := 3;
constant STATE_READ_ITERATE : integer := 4;
constant STATE_LOG_ITERATE : integer := 5;
constant STAGE_GET_READMUTEX : integer := 6;
constant STATE_WAIT_ROUTE_INDEX : integer := 7;

signal fsm_state : integer range 0 to 7 := STATE_INIT;

begin

dest <= dest_internal; -- concurrent assignment

process (clock) begin
	if rising_edge(clock) then
		if tick_time = '1' then
			curTime <= curTime + 1; -- may need to do some casting here
			comp_done <= '0';
		else	
			case fsm_state is
				when STATE_INIT =>
					mutex_write(route_index) <= '0'; -- set in STATE_WAITMUTEX
				
					if gen_task = '1' then -- if we need to generate a packet to send
						gen_ack <= '1'; 
						dest_internal <= gen_dest;
						src <= SELF_ADDR;	
						sent_time <= curTime; -- send time of packet is now!
						fsm_state <= STATE_DID_LOOKUP_ROUTE;
						
					elsif read_done = '0' then -- if gen_task = '0', check memory for packet propagation
						comp_done <= '0';
						readmutex_req <= '1';
						fsm_state <= STAGE_GET_READMUTEX;
					else
						comp_done <= '1';
						-- we're done for this virtual clock tick
					end if;
				when STAGE_GET_READMUTEX => -- acquire read mutex
					if readmutex_grant = '1' then
						read_en <= '1';
						fsm_state <= STATE_READ_ITERATE;
					end if;
				
				when STATE_READ_ITERATE => -- read_sig was set to '1' last cycle
					if read_ok = '1' then -- we read a packet from memory
						read_en <= '0';
						readmutex_req <= '0'; -- release memory mutex (prevents deadlock)
						
						-- if destination = self
						if read_data(7 downto 0) = SELF_ADDR then -- we received a packet!! logging?
							logging_en <= '1';
							logging_data <= read_data;
							fsm_state <= STATE_LOG_ITERATE;
						else -- destination is another router; forward the packet
							sent_time <= read_data(31 downto 16);
							src <= read_data(15 downto 8);
							dest_internal <= read_data(7 downto 0);							
							fsm_state <= STATE_DID_LOOKUP_ROUTE;								
						end if;
						
					elsif read_done = '1' then
						readmutex_req <= '0';
						read_en <= '0';
						fsm_state <= STATE_INIT;
					end if;
				when STATE_LOG_ITERATE => -- iterate until logging mechanism confirms
					if logging_ack = '1' then
						logging_en <= '0';
						fsm_state <= STATE_INIT; -- go back to STATE_INIT
					end if;
				
				when STATE_DID_LOOKUP_ROUTE => -- sending; we just looked up the outgoing route index
					gen_ack <= '0'; 
					packet_time <= std_logic_vector(unsigned(curTime) + delay(route_index * 8 + 7 downto route_index * 8)); 
					
					-- takes an extra cycle before route_index is correct
					fsm_state <= STATE_WAIT_ROUTE_INDEX;
					
				when STATE_WAIT_ROUTE_INDEX =>
					-- request mutex for the proper index
					mutex_req(route_index) <= '1';
					fsm_state <= STATE_WAITMUTEX;
					
				when STATE_WAITMUTEX =>
					if mutex_grant(route_index) = '1' then
						-- write packet to mutex'd memory
						mutex_write(route_index) <= '1'; -- REMEMBER to reset mutex_write
						mutex_data(route_index * 48 + 47 downto route_index * 48) <= 
														packet_time & sent_time & src & dest_internal;
						
						-- release mutex (will be detected next cycle)
						mutex_req(route_index) <= '0';					
					
						fsm_state <= STATE_INIT;		
					end if;
				when others => null;
			end case;
		end if;
	end if;

end process;

end Behavioral;
