import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.TreeMap;


public class Main 
{
	private static TreeMap <String, Router> routers = new TreeMap <String, Router>();
	private static TreeMap <Integer, Integer> flows = new TreeMap <Integer, Integer>();
	
	private static final int LOGGING_ROUTER_NONE = -1;
	private static int loggingRouter = LOGGING_ROUTER_NONE;
	
	public static void main(String[] args)
	{
		String top_filename = "top.txt";
		String flow_filename = "flows.txt";
		
		if (args.length > 0)
			top_filename = args[0];
		
		if (args.length > 1)
			flow_filename = args[1];
		
		System.out.println("Reading topology from file " + top_filename);		
		readTop(top_filename);
		
		System.out.println("Reading flows from file " + flow_filename);
		readFlows(flow_filename);
		
		System.out.println("Using hardcoded logging router; router #" + loggingRouter);
		
		generateRoutingTables();
		generateCoordinator();
		generateTop();
		
		
		System.out.println("Exiting...");
		System.exit(0);
	}
	
	// generate the top level file
	private static void generateTop() 
	{
		VHDL_Entity top = new VHDL_Entity();
		top.createBlank("top");
		top.addSignal(new VHDL_Signal("clock", "std_logic", SignalDirection.IN, null));
		
		// serial communication
		top.addSignal(new VHDL_Signal("rx", "std_logic", SignalDirection.IN, null));
		top.addSignal(new VHDL_Signal("tx", "std_logic", SignalDirection.OUT, null));
		
		// internal signals
		top.addSignal(new VHDL_Signal("gen_traffic", "std_logic", SignalDirection.INTERNAL, "'0'"));
		top.addSignal(new VHDL_Signal("comp_done", "std_logic", SignalDirection.INTERNAL, "'0'"));
		top.addSignal(new VHDL_Signal("milliseconds", "unsigned(15 downto 0)", 
					SignalDirection.INTERNAL, "(others => '0')"));
		
		// create global coordinator
		VHDL_Entity coord = new VHDL_Entity("out/coord.vhd");
		coord.parse();
		ArrayList <String> coord_generics = coord.getDefaultGenericMapping();
		ArrayList <String> coord_ios = coord.getDefaultIOMapping();
		
		// add all non-clock signals to top
		for (int x = 1; x < coord.ioSignals.size(); ++x)
		{
			VHDL_Signal s = coord.ioSignals.get(x);
			top.addSignal(new VHDL_Signal(s.name, s.type, SignalDirection.INTERNAL, s.initial));
		}
		
		top.addSignal(new VHDL_Signal("proceed_coord", "std_logic", SignalDirection.INTERNAL, null));
		coord_ios.set(1,"proceed_coord");
		
		top.addBehaviorLine("proceed <= proceed_coord and (not comp_done);\n");
		
		top.addInstance("i_coord", coord, coord_ios, coord_generics);
		
		// create router instances
		for (Router r : routers.values())
		{
			VHDL_Entity router = new VHDL_Entity("comp_ele.vhd");
			router.parse();
			router.printEntity("out/comp_ele.vhd");
			
			ArrayList <String> router_generics = new ArrayList <String>();
			router_generics.add("std_logic_vector(to_unsigned(" + r.name + ", 8))"); // self addr
			router_generics.add("" + r.neighbors.size());
			
			ArrayList <String> router_ios = router.getDefaultIOMapping();
			// ### (0)clock is global input
			// ### (1)gen_task and (2)gen_dest and (13)gen_ack come from traffic generator
			router_ios.set(1, "gen_task" + r.name);
			router_ios.set(2, "gen_dest" + r.name);
			router_ios.set(13, "gen_ack" + r.name);
			
			VHDL_Entity gen = new VHDL_Entity("traffic_gen.vhd");
			gen.parse();
			gen.printEntity("out/traffic_gen.vhd");
			
			ArrayList<String> gen_generics = new ArrayList<String>();
			
			int flowTo = flows.get(Integer.parseInt(r.name)).intValue();
			//System.out.println("router " + r.name + " has a flow to " + flowTo);
			
			gen_generics.add("std_logic_vector(to_unsigned(" + 
					flowTo + ",8))"); // destination
			ArrayList <String> gen_ios = gen.getDefaultIOMapping();
			gen_ios.set(2, "gen_ack" + r.name);
			top.addSignal(new VHDL_Signal("gen_ack" + r.name, gen.ioSignals.get(2).type, 
					SignalDirection.INTERNAL, gen.ioSignals.get(1).initial));
			gen_ios.set(3, "gen_traffic");
			gen_ios.set(4, "gen_task" + r.name);
			top.addSignal(new VHDL_Signal("gen_task" + r.name, gen.ioSignals.get(4).type, 
					SignalDirection.INTERNAL, gen.ioSignals.get(3).initial));
			gen_ios.set(5, "gen_dest" + r.name);
			top.addSignal(new VHDL_Signal("gen_dest" + r.name, gen.ioSignals.get(5).type, 
					SignalDirection.INTERNAL, gen.ioSignals.get(5).initial));
			
			top.addInstance("i_gen" + r.name, gen, gen_ios, gen_generics);
			
			// ### (3)read_done, (4)read_ok, (5)read_data come from mem_interface
			// ### so does (17)read_en
			router_ios.set(3, "read_done" + r.name);
			router_ios.set(4, "read_ok" + r.name);
			router_ios.set(5, "read_data" + r.name);
			router_ios.set(17, "read_en" + r.name);
			
			VHDL_Entity mem = new VHDL_Entity("mem_interface.vhd");
			mem.parse();
			mem.printEntity("out/mem_interface.vhd");
			
			// also copy block ram used by mem interface
			VHDL_Entity bram = new VHDL_Entity("my_bram.vhd");
			bram.parse();
			bram.printEntity("out/my_bram.vhd");
			
			ArrayList <String> mem_generics = new ArrayList <String>();
			ArrayList <String> mem_ios = mem.getDefaultIOMapping();
			mem_ios.set(1,"proceed"); // virtual_tick
			
			// add remaining per-router mem signals to top	
			for (int x = 2; x < mem.ioSignals.size(); ++x)
			{
				VHDL_Signal s = mem.ioSignals.get(x);
				String name = s.name + r.name;
				
				top.addSignal(new VHDL_Signal(name,s.type,SignalDirection.INTERNAL,s.initial));
				mem_ios.set(x, name);
			}
			
			top.addInstance("i_mem" + r.name, mem, mem_ios, mem_generics);
			
			// ### (6)tick_time is from global coordinator (proceed)
			router_ios.set(6, "proceed");
			
			// ### (7) logging_ack is hardwired to '1' for nonlogging routers
			// ### (18)loggin_en (19)logging_data
			if (Integer.parseInt(r.name) != loggingRouter)
			{
				router_ios.set(7,"'1'"); // logging_ack
				router_ios.set(18,"open"); // logging_en
				router_ios.set(19,"open"); // logging_data
			}
			else
			{
				// create logging instance
				VHDL_Entity logging = new VHDL_Entity("logging.vhd");
				logging.parse();
				logging.printEntity("out/logging.vhd");
				
				// also copy serial instance in output directory
				VHDL_Entity serial_bak = new VHDL_Entity("serial_bak.vhd");
				serial_bak.parse();
				serial_bak.printEntity("out/serial_bak.vhd");
				
				// add signals to top
				for (int x = 2; x <= 4; ++x)
				{
					VHDL_Signal s = logging.ioSignals.get(x);
					top.addSignal(new VHDL_Signal(s.name,s.type,SignalDirection.INTERNAL,s.initial));
				}
				
				top.addInstance("i_logging", logging, 
						logging.getDefaultIOMapping(), logging.getDefaultGenericMapping());
			}
			
			// ### (8) mutex_grant, (14)mutex_req, (15)mutex_write, (16)mutex_data
			// ### also (11)readmutex_grant and (12)readmutex_req 
			
			// create mutex internal signals
			// basically mutex_grant and mutex_req are for all connections including self
			// mutex_write and mutex_data are for all connections not including self
			top.addSignal(new VHDL_Signal("mutex_grant" + r.name + "_" + r.name, "std_logic",
					SignalDirection.INTERNAL, null));
			top.addSignal(new VHDL_Signal("mutex_req" + r.name + "_" + r.name, "std_logic",
					SignalDirection.INTERNAL, "'0'"));
			
			for (Router neb : r.neighbors)
			{
				top.addSignal(new VHDL_Signal("mutex_grant" + r.name + "_" + neb.name, "std_logic",
						SignalDirection.INTERNAL, null));
				top.addSignal(new VHDL_Signal("mutex_req" + r.name + "_" + neb.name, "std_logic",
						SignalDirection.INTERNAL, "'0'"));
				top.addSignal(new VHDL_Signal("mutex_write" + r.name + "_" + neb.name, "std_logic",
						SignalDirection.INTERNAL, "'0'"));
				top.addSignal(new VHDL_Signal("mutex_data" + r.name + "_" + neb.name, 
						"std_logic_vector(47 downto 0)", SignalDirection.INTERNAL, "(others => '0')"));
			}
			
			//create router_mutex signals and assign their behavior
			top.addSignal(new VHDL_Signal("router" + r.name + "_mutex_grant", "std_logic_vector(" 
					+ (r.neighbors.size() - 1) + " downto 0)", SignalDirection.INTERNAL, null));
			top.addSignal(new VHDL_Signal("router" + r.name + "_mutex_req", "std_logic_vector(" 
					+ (r.neighbors.size() - 1) + " downto 0)", SignalDirection.INTERNAL, null));
			top.addSignal(new VHDL_Signal("router" + r.name + "_mutex_write", "std_logic_vector(" 
					+ (r.neighbors.size() - 1) + " downto 0)", SignalDirection.INTERNAL, null));
			top.addSignal(new VHDL_Signal("router" + r.name + "_mutex_data", "std_logic_vector(48*" 
					+ (r.neighbors.size() - 1) + "+47 downto 0)", SignalDirection.INTERNAL, null));
			
			// router_mutex_grant needs to be aggregated
			String router_mutex_grant = "";
			
			boolean first = true;
			for (int x = r.neighbors.size() - 1; x >= 0; --x) // it's downto
			{
				Router neb = r.neighbors.get(x);
				
				if (first)
					first = false;
				else
					router_mutex_grant += " & ";
				
				router_mutex_grant += "mutex_grant" + neb.name + "_" + r.name;
			}
			
			if (r.neighbors.size() > 1)
				top.addBehaviorLine("router" + r.name + "_mutex_grant <= " + router_mutex_grant + ";");
			else
				top.addBehaviorLine("router" + r.name + "_mutex_grant(0) <= " + router_mutex_grant + ";");
				
			// req write and data need to be deaggregated
			for (int x = r.neighbors.size() - 1; x >= 0; --x) // it's downto
			{
				Router neb = r.neighbors.get(x);
				
				top.addBehaviorLine("mutex_req" + r.name + "_" + neb.name + " <= router" + r.name
						+ "_mutex_req(" + x + ");");
				
				top.addBehaviorLine("mutex_write" + r.name + "_" + neb.name + " <= router" + r.name
						+ "_mutex_write(" + x + ");");
				
				top.addBehaviorLine("mutex_data" + r.name + "_" + neb.name + " <= router" + r.name
						+ "_mutex_data(" + (48*x+47) + " downto " + (48*x) + ");");
			}
			
			// self (0_0) will get assigned by signal assignment to read_mutex
			
			// assign instances
			router_ios.set(8, "router" + r.name + "_mutex_grant");
			router_ios.set(14, "router" + r.name + "_mutex_req");
			router_ios.set(15, "router" + r.name + "_mutex_write");
			router_ios.set(16, "router" + r.name + "_mutex_data");
			
			router_ios.set(11, "mutex_grant" + r.name + "_" + r.name);
			router_ios.set(12, "mutex_req" + r.name + "_" + r.name);
			
			////////////////////////////////////////////
			// finally create mutex instance
			VHDL_Entity mutex = new VHDL_Entity("mutex.vhd");
			mutex.parse();
			mutex.printEntity("out/mutex.vhd");
			
			// create mutex aggregate signals for feeding into mutexes
			// req write and data are aggregated			
			String mutex_req = "";
			String mutex_write = "";
			String mutex_data = "";
			
			first = true;
			for (int x = r.neighbors.size() - 1; x >= 0; --x) // it's downto
			{
				Router neb = r.neighbors.get(x);
				
				if (first)
					first = false;
				else
				{
					mutex_req += " & ";
					mutex_write += " & ";
					mutex_data += " & ";
				}
				
				mutex_req += "mutex_req" + neb.name + "_" + r.name;
				mutex_write += "mutex_write" + neb.name + "_" + r.name;
				mutex_data += "mutex_data" + neb.name + "_" + r.name;
			}
			
			// add self
			mutex_req += " & " + "mutex_req" + r.name + "_" + r.name;
			mutex_write += " & \"0\""; // no mutex_write(self)
			mutex_data += " & x\"000000000000\"";  // no mutex_data(self)
			
			top.addSignal(new VHDL_Signal("mutex_req" + r.name, "std_logic_vector(" 
					+ (r.neighbors.size()) + " downto 0)",SignalDirection.INTERNAL, null));
			top.addSignal(new VHDL_Signal("mutex_write" + r.name, "std_logic_vector(" 
					+ (r.neighbors.size()) + " downto 0)",SignalDirection.INTERNAL, null));
			top.addSignal(new VHDL_Signal("mutex_data" + r.name,  "std_logic_vector(48*" 
					+ (r.neighbors.size()+1) + "-1 downto 0)",SignalDirection.INTERNAL, null));
			
			top.addBehaviorLine("mutex_req" + r.name + " <= " + mutex_req + ";");			
			top.addBehaviorLine("mutex_write" + r.name + " <= " + mutex_write + ";");			
			top.addBehaviorLine("mutex_data" + r.name + " <= " + mutex_data + ";");
			
			// grant is deaggregated
			top.addSignal(new VHDL_Signal("mutex_grant" + r.name, "std_logic_vector(" 
					+ (r.neighbors.size()) + " downto 0)",SignalDirection.INTERNAL, "(others => '0')"));
			
			for (int x = 0; x < r.neighbors.size(); ++x)
			{
				Router neb = r.neighbors.get(x);
				
				top.addBehaviorLine("mutex_grant" + r.name + "_" + neb.name + " <= mutex_grant" + r.name 
						+ "(" + (x+1) + ");");
			}
			
			// self
			top.addBehaviorLine("mutex_grant" + r.name + "_" + r.name + " <= mutex_grant" + r.name 
					+ "(0);");
			
			
			ArrayList <String> mutex_generics = new ArrayList <String>();
			mutex_generics.add("" + (r.neighbors.size() + 1)); // width (+1 for self)
			ArrayList <String> mutex_ios = mutex.getDefaultIOMapping();
			
			mutex_ios.set(1,"mutex_req" + r.name);
			mutex_ios.set(2,"mutex_write" + r.name);
			mutex_ios.set(3,"mutex_data" + r.name);
			mutex_ios.set(4,"mutex_grant" + r.name);
			mutex_ios.set(5,"write_en" + r.name); // write to mem
			mutex_ios.set(6,"write_data" + r.name); // data to mem
			
			top.addInstance("i_mutex" + r.name, mutex, mutex_ios, mutex_generics);
			
			// ### (9)delay (constant!!!)
			String delay = "unsigned(";
			first = true;
			
			for (int x = 0; x < r.neighbors.size(); ++x)
			{
				if (first)
					first = false;
				else
					delay += " & ";
				
				delay += "std_logic_vector(to_unsigned(" + r.neb_delays.get(x).intValue() + ",8))";
			}
			
			delay += ")";
			
			// we have to make this a signal b/c simulator is bugg'd if we use it directly
			top.addSignal(new VHDL_Signal("delay" + r.name, "unsigned(" + (8*r.neighbors.size()-1) + 
					" downto 0)", SignalDirection.INTERNAL, delay));
			
			router_ios.set(9, "delay" + r.name);
			
			// ### (10)route_index and (21)dest connected to routing table
			router_ios.set(10, "route_index" + r.name);
			router_ios.set(21, "dest" + r.name);
			
			VHDL_Entity route_table = new VHDL_Entity("out/route_table" + r.name + ".vhd");
			route_table.parse();
			
			ArrayList <String> route_table_generics = new ArrayList<String>();
			route_table_generics.add("" + r.neighbors.size()); // CONNECTIONS
			
			System.out.println("## router" + r.name + " has " + r.neighbors.size() + " nebs");
			
			
			ArrayList <String> route_table_ios = route_table.getDefaultIOMapping();
			route_table_ios.set(1,"dest" + r.name); // dest
			top.addSignal(new VHDL_Signal("dest" + r.name, "std_logic_vector(7 downto 0)", 
					SignalDirection.INTERNAL, "(others => '0')"));
			route_table_ios.set(2,"route_index" + r.name); // route_index
			top.addSignal(new VHDL_Signal("route_index" + r.name, 
					"integer range 0 to " + (r.neighbors.size() -1),
					SignalDirection.INTERNAL,"0"));
					
			top.addInstance("i_routetable" + r.name, route_table, route_table_ios, route_table_generics);
			
			// ### (20)comp_done connected to ready from global coordinator
			router_ios.set(20, "ready_" + r.name);
			
			top.addInstance("i_router" + r.name, router, router_ios, router_generics);
		}
		
		if (loggingRouter == LOGGING_ROUTER_NONE)
		{
			// create logging instance still for outputting time of simulation in ticks
			
			VHDL_Entity logging = new VHDL_Entity("logging.vhd");
			logging.parse();
			logging.printEntity("out/logging.vhd");
			
			// also copy serial instance in output directory
			VHDL_Entity serial_bak = new VHDL_Entity("serial_bak.vhd");
			serial_bak.parse();
			serial_bak.printEntity("out/serial_bak.vhd");
			
			// add signals to top
			for (int x = 2; x <= 4; ++x)
			{
				VHDL_Signal s = logging.ioSignals.get(x);
				top.addSignal(new VHDL_Signal(s.name,s.type,SignalDirection.INTERNAL,s.initial));
			}
			
			top.addInstance("i_logging", logging, 
					logging.getDefaultIOMapping(), logging.getDefaultGenericMapping());
		}
		
		top.addBehavior("top_beh.txt");
		
		top.printEntity("out/top.vhd");
	}
	
	// generate the global coordinator
	private static void generateCoordinator() 
	{
		try
		{
			VHDL_Entity ent = new VHDL_Entity();
			ent.createBlank("coordinator");
			ent.addSignal(new VHDL_Signal("clock", "std_logic", SignalDirection.IN, null));
			ent.addSignal(new VHDL_Signal("proceed", "std_logic", SignalDirection.OUT, null));
			ent.addSignal(new VHDL_Signal("reset_proceed", "std_logic", SignalDirection.INTERNAL, "'0'"));
			
			String beh_file = "out/global_beh.txt";
			PrintWriter fout = new PrintWriter(new BufferedWriter(
					new FileWriter(beh_file)));
			
			fout.println("process(clock) begin");
			fout.println("\tif rising_edge(clock) then");
			
			fout.print("\t\tif reset_proceed = '0'");
		
			for (Router r : routers.values())
			{
				fout.print(" and ");
				
				fout.print("ready_" + r.name + " = '1'");
				ent.addSignal(new VHDL_Signal("ready_" + r.name, "std_logic", SignalDirection.IN, null));
			}
			
			fout.println(" then");
			
			fout.println("\t\t\tproceed <= '1';");
			fout.println("\t\t\treset_proceed <= '1';");
			fout.println("\t\telse");
			fout.println("\t\t\treset_proceed <= '0';");
			fout.println("\t\t\tproceed <= '0';");
			
			fout.println("\t\tend if;");
			
			fout.println("\tend if;");
			fout.println("end process;");
			
			fout.close();
			
			ent.addBehavior(beh_file);			
			ent.printEntity("out/coord.vhd");		
		}
		catch (IOException e)
		{
			System.out.println(e);
		}
	}

	private static void generateRoutingTables() 
	{
		for (Router source : routers.values())
		{
			for (Router r : routers.values())
				r.reset();
						
			source.cost = 0;
			source.DFS();
			String min = null, max = null;
			Router curParent = null;
			System.out.println("\n--- Route Table for " + source.name + " ---");
			
			try
			{
				String beh_file = "out/router" + source.name + "_beh.txt";
				PrintWriter fout = new PrintWriter(new BufferedWriter(
						new FileWriter(beh_file)));
				
				fout.println("process(clock) begin");
				fout.println("\tif rising_edge(clock) then");
				
				boolean first = true;
			
				TreeMap <Integer, Router> sortedRouters = new TreeMap <Integer, Router>(); 
				
				for (String s : routers.keySet())
				{
					Router r = routers.get(s);
					
					sortedRouters.put(Integer.parseInt(s), r);
				}
				
				for (Router r : sortedRouters.values())
				{				
					if (!r.equals(source))
					{
						if (min == null)
						{
							min = max = r.name;
							curParent = r.findFirstParent(); 
						}
						else
						{
							Router myParent = r.findFirstParent();
							
							if (myParent.equals(curParent))
								max = r.name;
							else
							{
								System.out.println("[" + min + ", " + max + "] routed through " + curParent.name);
								
								fout.print("\t\t");
								
								if (!first)
									fout.print("els");
								
								if (min != max)
									fout.println("if unsigned(dest) >= " + min + " and unsigned(dest) <= " + max + " then ");
								else
									fout.println("if unsigned(dest) = " + min + " then ");
								
								String neb_name = curParent.name;
								int index = 0;
								
								for (; index < source.neighbors.size(); ++index)
								{
									if (source.neighbors.get(index).name.equals(neb_name))
										break;
								}
																
								fout.println("\t\t\troute_index <= " + index + ";");
								
								
								min = max = r.name;
								curParent = myParent;
								first = false;
							}
						}
					}
					else
						System.out.println("self <- " + r.name);
				}
				
				System.out.println("[" + min + ", " + max + "] routed through " + curParent.name);
				
				fout.print("\t\t");
				
				if (!first)
					fout.print("els");
				
				if (min != max)
					fout.println("if unsigned(dest) >= " + min + " and unsigned(dest) <= " + max + " then ");
				else
					fout.println("if unsigned(dest) = " + min + " then ");
				
				String neb_name = curParent.name;
				int index = 0;
				
				for (; index < source.neighbors.size(); ++index)
				{
					if (source.neighbors.get(index).name.equals(neb_name))
						break;
				}
				
				fout.println("\t\t\troute_index <= " + index + ";");
				
				fout.println("\t\tend if;");
				fout.println("\tend if;");
				fout.println("end process;");
				
				fout.close();
				
				// create the VHDL file
				String router_name = "route_table" + source.name;
				VHDL_Entity r = new VHDL_Entity("route_table.vhd");
				r.parseReplace("route_table",router_name);
				r.addBehavior(beh_file);				
				r.printEntity("out/" + router_name + ".vhd");
				
			}
			catch (IOException e)
			{
				System.out.println(e);
			}
		}
	}
	
	private static void readFlows(String filename)
	{
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(filename));
			
			for (String line = in.readLine(); line != null; line = in.readLine())
			{
				String[] nodes = line.split(" ");
				
				if (nodes.length == 0) // blank line
					break;
				
				Integer a = Integer.parseInt(nodes[0]);
				Integer b = Integer.parseInt(nodes[1]);
				
				flows.put(a, b);				
			}
			
			in.close();
		} 
		catch (IOException e)
		{
			System.out.println(e);
		}
	}

	private static void readTop(String filename) 
	{
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(filename));
			
			for (String line = in.readLine(); line != null; line = in.readLine())
			{
				String[] nodes = line.split(" ");
				
				if (nodes.length == 0) // blank line
					break;
				
				// assumption: nodes.size = 3
				Router r1 = routers.get(nodes[0]);
				
				if (r1 == null)
				{
					r1 = new Router(nodes[0]);
					routers.put(nodes[0],r1);
				}
				
				Router r2 = routers.get(nodes[1]);
				
				if (r2 == null)
				{
					r2 = new Router(nodes[1]);
					routers.put(nodes[1],r2);
				}
				
				System.out.println("r1 = " + r1 + "; r2 = " + r2);
				
				// manipulate r1 and r2
				r1.neighbors.add(r2);
				r1.neb_delays.add(new Integer(Integer.parseInt(nodes[2])));
				
				// links are uni-directional (listed again for reverse direction)
			}
			
			in.close();
		} 
		catch (IOException e)
		{
			System.out.println(e);
		}
	}
}
