-- Stan Bak
-- Logging for network simulator - 12-08

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;
use work.bak.all;

entity logging is
	generic	
	(
		 G_CLOCK_RATE : integer := 33000000; -- 33 mhz clock rate
		 G_BAUD_RATE  : integer := 115200
	);
	port 
	( 
		clock			: in std_logic;
		comp_done : in std_logic; -- global computation done?
		logging_en	: in std_logic;
		logging_data : in std_logic_vector(47 downto 0);
		
		
		logging_ack : out std_logic := '0';
		
		-- serial port signals
		rx : in std_logic;
		tx : out std_logic
	); 
end logging;

architecture Behavioral of logging is

	component serial_bak is
	generic
	(
		CLOCK_RATE : integer;    	-- hertz
		BAUD_RATE  : integer  		-- bits/sec
	);	
	port 
	( 
			send_now	     	: in 	 std_logic;
			clock      		: in    std_logic; 
         rx           	: in    std_logic; 
         send_data     	: in    std_logic_vector (7 downto 0); 
         data_present 	: out   std_logic; 
         receive_data   : out   std_logic_vector (7 downto 0); 
         tx           	: out   std_logic
	);
end component;

 -- serial port signals
	 signal send_now : std_logic := '0';
	 signal send_data : std_logic_vector(7 downto 0);
	 
	 -- 33mhz / 115200 baud * 10 bits per byte(8 + start bit + stop bit) + gamma
	 constant INTERSEND_TICKS_RESET : integer := (10 * G_CLOCK_RATE / G_BAUD_RATE) + 25; 
	 constant STATE_CR : integer := 12;
	 constant STATE_LF : integer := 13;
	 constant STATE_WAIT : integer := 14;
	 constant STATE_TICKS : integer := 15; -- print ticks at the end of the computation
	 
	 signal data : std_logic_vector(47 downto 0);
	 signal send_state : integer range 0 to 15 := STATE_CR; -- first 11 are nibbles then cr then lf then wait
	 -- first printed character is newline
	 signal intersend_ticks : integer range 0 to INTERSEND_TICKS_RESET := 0;
	 signal ticks : unsigned(31 downto 0) := (others => '0');
	 
	 -- the state of printing the ticks when computation ends
	 constant PRINT_TICKS_STATE_DONE : integer := 15;
	 signal print_ticks_state : integer range 0 to 15 := 0;
	
begin

	i_serial_bak : serial_bak 
	 generic map
	 (
		 CLOCK_RATE=>G_CLOCK_RATE, -- 33 mhz clock rate
		 BAUD_RATE=>G_BAUD_RATE
	 )	
	 port map
	 (
		 clock => clock,
		 send_now => send_now,
		 send_data => send_data,
		 data_present => open,
		 receive_data => open,
		 rx => rx,
		 tx => tx		
	 );

process (clock) 

variable nibble : std_logic_vector(3 downto 0);

begin
	if rising_edge(clock) then
		if comp_done <= '0' then
			ticks <= ticks + 1;
		end if;
	
		if intersend_ticks > 0 then
			intersend_ticks <= intersend_ticks - 1;
			logging_ack <= '0';
			send_now <= '0';
		else
			if send_state = STATE_WAIT then				
				if comp_done = '1' and print_ticks_state = 0 then -- output number of ticks
					send_state <= STATE_TICKS;
				elsif logging_en = '1' then
					data <= logging_data;
					send_state <= 0;
				end if;			
			else
				if send_state <= 11 then -- send a nibble of data				
					
					-- the nibble we want to send is at offset (47-4*send_state downto 44-4*send_state)
					nibble := data(47-(4*send_state) downto 44-(4*send_state));
					
					send_data <= unsigned_to_hexstring_c(unsigned(nibble));					
					
				elsif send_state = STATE_CR then
					send_data <= x"0d";
				
				elsif send_state = STATE_LF then
					send_data <= x"0a";
					logging_ack <= '1';
				elsif send_state = STATE_TICKS then
					if print_ticks_state < 8 then
						nibble := std_logic_vector(ticks(31-(4*print_ticks_state) downto 28-(4*print_ticks_state)));
						send_data <= unsigned_to_hexstring_c(unsigned(nibble));		
						print_ticks_state <= print_ticks_state + 1;
					else
						print_ticks_state <= PRINT_TICKS_STATE_DONE;
						send_state <= STATE_WAIT;
					end if;
					
				end if;
				
				if send_state /= STATE_TICKS then
					send_state <= send_state + 1;
				end if;
				
				send_now <= '1';
				intersend_ticks <= INTERSEND_TICKS_RESET;
			end if;
		end if;
	
	end if;
end process;

end Behavioral;

