-- Stan Bak
-- Memory Interface for network simulator - 12-08
-- the signals in this module are assumed to be mutexed correctly

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;
use work.bak.all;

entity mem_interface is
	port 
	( 
		clock			: in std_logic;
		virtual_tick : in std_logic; -- tick virtual clock?
		write_en : in std_logic;
		write_data : in std_logic_vector (47 downto 0);
		read_en : in std_logic;
		
		read_done : out std_logic := '0'; -- done reading for this virtual tick?
		read_ok : out std_logic := '0'; -- data populated?
		read_data : out std_logic_vector(47 downto 0) := (others => '0')
	); 
end mem_interface;

architecture Behavioral of mem_interface is

	component my_bram is
generic(
	ADDR_WIDTH : integer := 8;
	DATA_WIDTH : integer := 48
);
port(
	clock : in std_logic;
	addr : in std_logic_vector(ADDR_WIDTH - 1 downto 0);
	data_in : in std_logic_vector(DATA_WIDTH - 1 downto 0);
	write_enable : in std_logic;
	
	data_out : out std_logic_vector(DATA_WIDTH - 1 downto 0)
);
end component;

constant MAX_ADDR : integer := 255;
constant ADDR_LEN : integer := 8;

signal write_enable : std_logic := '0';
signal addr : std_logic_vector(7 downto 0) := (others => '0');
signal data_in : std_logic_vector(47 downto 0) := (others => '0');
signal data_out : std_logic_vector(47 downto 0) := (others => '0');

-- one past the last valid address with packet data
signal past_last_valid_addr : integer range 0 to MAX_ADDR := 0;

-- the current address in the virtual tick while we iterate
signal cur_addr : integer range 0 to MAX_ADDR := 0;

signal mill : unsigned (15 downto 0) := (others => '0');


constant STATE_READ_ADDR : integer := 0;
constant STATE_COMPARE_TIME : integer := 1;
constant STATE_READ_LAST : integer := 2;
constant STATE_READ_WAIT : integer := 3;
constant STATE_READ_LAST_WAIT : integer := 4;
signal read_fsm_state : integer range 0 to 4 := STATE_READ_ADDR;
signal forward_packet : std_logic_vector(47 downto 0) := (others => '0');

begin



bram_i : my_bram
generic map(
	ADDR_WIDTH => ADDR_LEN,
	DATA_WIDTH => 48
)
port map(
	clock => clock,
	addr => addr,
	data_in => data_in,
	write_enable => write_enable,
	
	data_out => data_out
);

process (clock) 

variable next_addr : integer range 0 to MAX_ADDR;
variable forward_time : std_logic_vector(15 downto 0);

begin
	if rising_edge(clock) then
		if virtual_tick = '1' then
			read_ok <= '0';
			
			if past_last_valid_addr > 0 then
				read_done <= '0';
			else
				read_done <= '1';
			end if;
			
			cur_addr <= 0;
			write_enable <= '0';
			mill <= mill + 1;
			read_fsm_state <= STATE_READ_ADDR;
		elsif write_en = '1' then
			read_ok <= '0';
			write_enable <= '1'; -- REMEMBER to disable write_enable in all potential transitions
			data_in <= write_data;
			
			if past_last_valid_addr < MAX_ADDR then
				next_addr := past_last_valid_addr + 1;
				read_done <= '0'; -- we're not done reading if we just wrote something
			else
				next_addr := MAX_ADDR;
			end if;
			
			addr <= 	std_logic_vector(to_unsigned(past_last_valid_addr,8));
			past_last_valid_addr <= next_addr;
		else
		
			-- if we are iterating through packets
			if read_en = '1' and cur_addr < past_last_valid_addr then
							
				case read_fsm_state is
					when STATE_READ_ADDR =>
						read_ok <= '0';
						write_enable <= '0';
						addr <= std_logic_vector(to_unsigned(cur_addr,8));
						read_fsm_state <= STATE_READ_WAIT;			
						forward_packet <= (others => '0');
					when STATE_READ_WAIT =>
						-- extra cycle of latency for read memory to take effect
						write_enable <= '0';
						read_ok <= '0';
						read_fsm_state <= STATE_COMPARE_TIME;
					when STATE_COMPARE_TIME =>
						write_enable <= '0';
						read_ok <= '0';
						-- check if time in data is less than or equal to current time
						forward_time := data_out(47 downto 32);
						
						if unsigned(forward_time) <= mill then -- propagate the packet
							forward_packet <= data_out; -- but don't set read_ok yet, hold onto mutex for swap
							
							-- swap out the data
							addr <= std_logic_vector(to_unsigned(past_last_valid_addr,8) - 1);
							read_fsm_state <= STATE_READ_LAST_WAIT;
							
						else						
							cur_addr <= cur_addr + 1;
							
							if cur_addr + 1 >= past_last_valid_addr then
								read_done <= '1';
							end if;
						
							read_fsm_state <= STATE_READ_ADDR;
						end if;
					when STATE_READ_LAST_WAIT => 
						-- extra cycle of read latency
						write_enable <= '0';
						read_ok <= '0';
						read_fsm_state <= STATE_READ_LAST;
					when STATE_READ_LAST =>
						-- we just read the last data, put it into memory location cur_addr
						addr <= std_logic_vector(to_unsigned(cur_addr,8));
						data_in <= data_out;
						write_enable <= '1'; -- REMEMBER to reset this on all paths
						read_data <= forward_packet;
						read_ok <= '1';
						
						if past_last_valid_addr > 0 then
							past_last_valid_addr <= past_last_valid_addr - 1;
						end if;
												
						if cur_addr >= past_last_valid_addr - 1 or past_last_valid_addr = 0 then
							read_done <= '1'; -- this is ok since read_ok takes preference over read_done in comp_ele
						end if;
						
						read_fsm_state <= STATE_READ_ADDR;
					
					when others =>
				end case;
			
			else
				if read_en = '1' then
					read_done <= '1';
				end if;
				
				write_enable <= '0';
				read_ok <= '0';
			end if;
		
		end if;
	
	end if;
end process;

end Behavioral;

