--- Stan Bak
--- 6-08
--- Serial I/O Module

-- You can copy and paste this component defintion
--component serial_bak is
-- generic
--	(
--		CLOCK_RATE : integer;    	-- hertz
--		BAUD_RATE  : integer  		-- bits/sec
--	);	
--	port 
--	( 
--			send_now	     	: in 	 std_logic;
--			clock      		: in    std_logic; 
--         rx           	: in    std_logic; 
--         send_data     	: in    std_logic_vector (7 downto 0); 
--         data_present 	: out   std_logic; 
--         receive_data    	: out   std_logic_vector (7 downto 0); 
--         tx           	: out   std_logic
--	);
--end component;

-- You can copy and paste this component instantiation template
--i_serial_bak : serial_bak 
-- generic map
-- (
--	 CLOCK_RATE=>33000000, -- 33 mhz clock rate
--	 BAUD_RATE=>9600
-- )	
-- port map
-- (
--	 clock => ,
--	 send_now => ,
--	 send_data => ,
--	 data_present => ,
--	 receive_data => ,
--	 rx => rx,
--	 tx => tx		
-- );

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity serial_bak is
	generic
	(
		CLOCK_RATE : integer := 33000000;    	-- hertz
		BAUD_RATE  : integer := 9600  			-- bits/sec
	);	   
   port 
	( 
		send_now			: in 	 std_logic;
		clock        	: in    std_logic; 
      rx           	: in    std_logic; 
      send_data     	: in    std_logic_vector (7 downto 0); 
      data_present 	: out   std_logic; 
      receive_data   : out   std_logic_vector (7 downto 0); 
      tx           	: out   std_logic
	);
end serial_bak;

architecture Behavioral of serial_bak is

-- sending
constant CLOCKS_PER_BIT : integer := CLOCK_RATE/BAUD_RATE;
constant IDLE_SIGNAL : std_logic := '1';

signal my_send_data : std_logic_vector (7 downto 0); -- first and last are start and stop bits, lsb first
signal bit_to_send : integer range 0 to 7 := 0;
signal counter_send : integer := 0;

constant STATE_IDLE : integer := 0;
constant STATE_STARTBIT : integer := 1;
constant STATE_DATA : integer := 2;
constant STATE_STOPBIT : integer := 3;
signal sending_fsm_state : integer range 0 to 3 := STATE_IDLE;

-- receiving
constant RECIEVE_SAMPLE_TICKS : integer := CLOCKS_PER_BIT/16;
constant NUM_ONES_INIT : integer := 7;
signal num_ones : integer range 0 to 7 := NUM_ONES_INIT;
signal counter_receive : integer := 0;
signal bit_to_receive : integer range 0 to 7 := 0;
signal receive_start : std_logic := '0';
signal next_sample : integer range 0 to 31;
signal reset_num_ones : std_logic := '0';
signal ignore_ticks : integer range 0 to 15 := 0;

begin

p_counticks : process (clock) begin
	if rising_edge(clock) then
	
		-----------------------------------
		------------- RECEIVE -------------
		-----------------------------------	
		if counter_receive < RECIEVE_SAMPLE_TICKS then
			counter_receive <= counter_receive + 1;
			data_present <= '0';
			
			-- after we receive a byte reset the number of ones so we dont start up again right away
			if reset_num_ones = '1' then
				reset_num_ones <= '0';
				num_ones <= NUM_ONES_INIT;
			end if;
		else
			counter_receive <= 0;
			
			-- num_ones is near 0 if we think the current value is a '0', and near 7 if we think '1'
			-- this acts as a sort of filter for noise
			if rx = '1' and num_ones < 7 then
				num_ones <= num_ones + 1;
			elsif rx = '0' and num_ones > 0 then
				num_ones <= num_ones - 1;
			end if;
			
			-- we ignore start conditions for 16 samples after we receive the last bit (so the stop bit)
			if ignore_ticks > 0 then
				ignore_ticks <= ignore_ticks - 1;
			end if;
			
			if receive_start = '0' then
				
				if num_ones = 0 and ignore_ticks = 0 then -- start!
					receive_start <= '1';
					
					-- we want to be 3/4 into the next sample, we're already 1/2 way in
					--16 (one sample) + 4 (1/4 sample) = 20 ticks
					next_sample <= 20; 
					bit_to_receive <= 0;
				end if;
				
			else -- receive_start = '1'
				if next_sample > 0 then
					next_sample <= next_sample - 1;
				else
					next_sample <= 15;
					
					-- use the best guess for the current value
					if num_ones <= 3 then 
						receive_data(bit_to_receive) <= '0';			
					else
						receive_data(bit_to_receive) <= '1';
					end if;
					
					if bit_to_receive = 7 then
						receive_start <= '0';
						data_present <= '1';
						reset_num_ones <= '1';
						ignore_ticks <= 15; -- do not start for one sample period
					else
						bit_to_receive <= bit_to_receive + 1;
					end if;
				end if;	
			end if;
			
		end if;
	
		----------------------------------
		-------------- SEND --------------
		----------------------------------
		if sending_fsm_state = STATE_IDLE and send_now = '1' then
			my_send_data <= send_data;
			sending_fsm_state <= STATE_STARTBIT;
		elsif counter_send < CLOCKS_PER_BIT then
			counter_send <= counter_send + 1;
		else
			counter_send <= 0;
			
			if sending_fsm_state = STATE_STARTBIT then
				tx <= '0';
				sending_fsm_state <= STATE_DATA;
				bit_to_send <= 0;
			elsif sending_fsm_state = STATE_DATA then
				tx <= my_send_data(bit_to_send);
				
				if bit_to_send = 7 then
					sending_fsm_state <= STATE_STOPBIT;
				else
					bit_to_send <= bit_to_send + 1;
				end if;				
			elsif sending_fsm_state = STATE_STOPBIT then
				tx <= '1';
				sending_fsm_state <= STATE_IDLE;
			else
				tx <= IDLE_SIGNAL;
			end if;
			
		end if;
	end if;
end process;


end Behavioral;

