/*
Stan Bak
Converts from tiers format to ns2 format
using stdin for input, stdout for output, stderr for info

This also creates a top.txt and flows.txt file which can be fed in to the FPGA-based network simulator

*/



#include <stdio.h>
#include <string>
#include <vector>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <sstream>
using namespace std;

struct Router
{
  Router() { flowTo = -1; }
  int x, y;
  vector <int> forward;

  int flowTo; // flow originating here
};

vector <Router> routers;

// adds if not present
int getRouterIndex(int x, int y)
{
  int rv = -1;
  
  for (int i = 0; i < (int)routers.size(); ++i)
  {
    if (routers[i].x == x && routers[i].y == y)
    {
      rv = i;
      break;
    }
  }

  if (rv == -1)
  {
    Router r;
    r.x = x;
    r.y = y;

    routers.push_back(r);
    rv = routers.size() - 1;
  }

  return rv;
}

void addLink(int index1, int index2)
{
  // add forward link
  bool found = false;
  for (int x = 0; x < (int)routers[index1].forward.size(); ++x)
  {
    int i = routers[index1].forward[x];

    if (i == index2)
    {
      found = true;
      break;
    }
  }

  if (!found)
    routers[index1].forward.push_back(index2);

  // add reverse route (maybe unnecessary)
  found = false;
  for (int x = 0; x < (int)routers[index2].forward.size(); ++x)
  {
    int i = routers[index2].forward[x];

    if (i == index1)
    {
      found = true;
      break;
    }
  }

  if (!found)
    routers[index2].forward.push_back(index1);
}

int getLatency(Router* r, Router* r2)
{
  int dx = r->x - r2->x;
  int dy = r->y - r2->y;
  int dist = (int)sqrt(dx * dx + dy * dy);
  int ms = dist/100; // link latency

  if (ms < 1)
    ms = 1;

  return ms;
}

void createNs2(int logRouter)
{
    // output ns2 file
  printf("set ns [new Simulator]\n");
  printf("\n");
  printf("set nf [open tmp/trace.txt w]\n");
  printf("\n");
  printf("proc finish {} {\n");
  printf("  global ns nf\n");
  printf("  $ns flush-trace\n");
  printf("  close $nf\n");
  printf("  exit 0\n");
  printf("}\n");
  printf("\n");

  // create routers
  for (int x = 0; x < (int)routers.size(); ++x)
    printf("set n(%i) [$ns node]\n",x);

  printf("\n");

  // create links
  for (int x = 0; x < (int)routers.size(); ++x)
  {
    Router* r = &routers[x];

    for (int y = 0; y < (int)r->forward.size(); ++y)
    {
      Router* r2 = &routers[r->forward[y]];
      int ms = getLatency(r,r2);

      printf("$ns simplex-link $n(%i) $n(%i) 100Mb %ims DropTail\n",x,r->forward[y], ms);
    }
  }

  printf("\n");

  // create flows
  for (int x = 0; x < (int)routers.size(); ++x)
  {  
    printf("set udp(%i) [new Agent/UDP]\n",x);
    printf("$ns attach-agent $n(%i) $udp(%i)\n",x,x);
    printf("set null(%i) [new Agent/Null]\n",x);
    printf("$ns attach-agent $n(%i) $null(%i)\n",routers[x].flowTo, x);
    printf("$ns connect $udp(%i) $null(%i)\n",x,x);
    printf("$udp(%i) set fid_ %i\n",x, 5000+x); // flow id??

    printf("set cbr(%i) [new Application/Traffic/CBR]\n",x);
    printf("$cbr(%i) attach-agent $udp(%i)\n",x,x);
    printf("$cbr(%i) set type_ CBR\n",x);
    printf("$cbr(%i) set packet_size_ 1000\n",x);
    printf("$cbr(%i) set rate_ 8.0Kb\n",x);
    printf("$cbr(%i) set random_ false\n",x);
  
    printf("$ns at 1.0 \"$cbr(%i) start\"\n",x);
    printf("$ns at 51.0 \"$cbr(%i) stop\"\n",x);
    printf("\n");
  }

  printf("\n");
  printf("$ns at 55.0 \"finish\"\n");
  
  printf("\n");

  // monitor incoming connections to logRouter
  Router* r = &routers[logRouter];
  for (int x = 0; x < (int)r->forward.size(); ++x)
    printf("$ns namtrace-queue $n(%i) $n(%i) $nf\n", r->forward[x], logRouter);

  printf("\n");
  printf("$ns run\n");
  printf("\n");
}

void createFPGA_Files()
{
  // output top.txt
  FILE* f = fopen("top.txt","w");

  if (!f)
    perror("fopen");
  else
  {
    for (unsigned int x = 0; x < routers.size(); ++x)
    {
      Router* r = &routers[x];

      for (unsigned int y = 0; y < r->forward.size(); ++y)
      {
	Router* r2 = &routers[r->forward[y]];
	int ms = getLatency(r,r2);

	fprintf(f,"%i %i %i\n",x, r->forward[y], ms);
      }
    }

    fclose(f);
  }

  // output flows.txt
  f = fopen("flows.txt","w");

  if (!f)
    perror("fopen");
  else
  {
    for (unsigned int x = 0; x < routers.size(); ++x)
    {
      Router* r = &routers[x];

      fprintf(f,"%i %i\n",x,r->flowTo);
    }

    fclose(f);
  }
}

int main(int argc, char** argv)
{
  // use random seed if provided
  if (argc > 1)
    srand(atoi(argv[1]));
  else // otherwise use time as random seed
    srand(time(0)); 

  string s;

  while (getline(cin,s))
  {
    if (strstr(s.c_str(),"# WAN "))
    {
      break;
    }
  }

  int x1,y1,x2,y2;
  bool source = true;

  // start reading
  while (getline(cin,s))
  {
    if (s.length() == 0 || s[0] == '#' || s[0] == 'e')
      continue;
    else if (s[0] == 'p') // pause (end of input)
      break;
    else
    {
      istringstream ss(s);

      if (source)
	ss >> x1 >> y1;
      else
      {
	ss >> x2 >> y2;
	
	int index1 = getRouterIndex(x1, y1);
	int index2 = getRouterIndex(x2, y2);

	addLink(index1,index2);
      }

      source = !source;
    }
  }

  // create random flows
  for (int x = 0; x < (int)routers.size(); ++x)
  {
    int dest = rand() % (routers.size()-1);
    
    // no self-flows!
    if (dest == x)
      dest = routers.size() - 1;

    routers[x].flowTo = dest;
  }

  // accumulate statistics
  int max = -1;
  int numUniEdge = 0;
  int minDistSq = 999999999;
  int maxDistSq = 0;
  int maxRouter = -1;
  int minDestNotZero = 9999;
  int minDestNotZeroRouter = -1;
  int minDestNotZeroNumNebs = 999;

  for (int x = 0; x < (int)routers.size(); ++x)
  {
    int r = routers[x].forward.size();
    
    for (int y = 0; y < (int)routers[x].forward.size(); ++y)
    {
      int dx = routers[x].x - routers[routers[x].forward[y]].x;
      int dy = routers[x].y - routers[routers[x].forward[y]].y;

      int distSq = dx * dx + dy * dy;

      if (distSq < minDistSq)
	minDistSq = distSq;

      if (distSq > maxDistSq)
	maxDistSq = distSq;
    }

    if (r > max)
    {
      maxRouter = x;
      max = r;
    }
      
    if (r == 1)
    {
      ++numUniEdge;
    }

    // compute # flow destinations that go to router x
    int dest = 0;

    for (int y = 0; y < (int)routers.size(); ++y)
    {
      if (routers[y].flowTo == x)
	++dest;
    }

    if (dest > 0 && dest <= minDestNotZero)
    {
      if (minDestNotZeroNumNebs > (int)routers[x].forward.size())
      {
	minDestNotZero = dest;
	minDestNotZeroRouter = x;
	minDestNotZeroNumNebs = routers[x].forward.size();
      }
    }
  }

  cerr << "total routers: " << routers.size() << endl;
  cerr << "max neighbors: " << max << " at router " << maxRouter << endl;
  cerr << "num one edge:  " << numUniEdge << endl;
  cerr << "dist range:    " << sqrt(minDistSq) << " -> " << sqrt(maxDistSq) << endl;
  cerr << "log router(" << minDestNotZero << " incoming): " << minDestNotZeroRouter << endl << endl;

  createNs2(minDestNotZeroRouter);
  createFPGA_Files();
    
  return 0;
}


